// toShare rest.js
///<reference path="user.js"/>
///<reference path="inoutboard.js"/>
'use strict';

const restVer = '/v0';
const basePath = location.origin + restVer;

/**
 * string from datetime
 * @param {Date} date 
 * @param {string} format
 * - yyyy=year(4 digit)
 * - MM=month(2 digit) M=month(variable digit)
 * - dd=day(2 digit) d=day(variable digit)
 * - HH=hour(2 digit) mm=minute(2 digit) ss=second(2 digit) SSS=millisecond(3 digit)
 * - week=weekday(漢字一文字)
 * @returns {string}
 * @description
 */
function formatDate(date, format){
	const week = ['日','月','火','水','木','金','土'];
	format = format.replace(/yyyy/g, date.getFullYear());
	format = format.replace(/MM/g, ('0' + (date.getMonth() + 1)).slice(-2));
	format = format.replace(/dd/g, ('0' + date.getDate()).slice(-2));
	format = format.replace(/M/g, ('' + (date.getMonth() + 1)).slice(-2));
	format = format.replace(/d/g, ('' + date.getDate()).slice(-2));
	format = format.replace(/HH/g, ('0' + date.getHours()).slice(-2));
	format = format.replace(/mm/g, ('0' + date.getMinutes()).slice(-2));
	format = format.replace(/ss/g, ('0' + date.getSeconds()).slice(-2));
	format = format.replace(/SSS/g, ('00' + date.getMilliseconds()).slice(-3));
	format = format.replace(/week/g, week[date.getDay()]);
	return format;
};

/**
 * ?? Nullish coalescing
 * @param {any} target 
 * @param {any} defaultValue 
 * @returns {any}
 */
function altHCO(target, defaultValue){
	// 2020 Nullish coalescing operator ??
	return ((target != null)||(target != undefined)) ? target : defaultValue;
}

/** toShare API Base Classes */
class restClient {
	/**
	 * @param {string} url - endpoint
	 * @param {string} jsonRoot 
	 */
	constructor(url, jsonRoot){
		if(!url){throw Error(`${this.constructor.name}.restClient: url is a required argument.`);}
		this.jsonRoot = jsonRoot ? jsonRoot : '';
		this.url = url;
		this.baseURL = this.url;
		/** An excerpt of json obtained by restGet */
		this.data = {};
	}
	/**
	 * @private
	 */
	async restFetch(initOpt){
		const response = await fetch(this.url,initOpt);
		if (!response.ok){
			throw Error(`${this.constructor.name}.restFetch: Error response: ${response.status} ${response.statusText}`);
		}
		return response;
	}
	/** If the root of the fetched json is this.jsonRoot, keep it in this.data */
	async restGet(){
		const response = await this.restFetch({
			method: 'GET',
			headers: {'Accept': 'application/json'}
		});
		const contentType = await response.headers.get('content-type');
		if (!contentType || !contentType.includes('application/json')){
			throw new TypeError(`${this.constructor.name}.restGet: mime not json`);
		}
		const json = await response.json();
		if(!json){throw Error(`${this.constructor.name}.restGet: json schema error`);}
		this.data = json[this.jsonRoot];
		return response;
	}
	async restPost(jsondata){
		if(!jsondata){throw Error(`${this.constructor.name}.restPost: unknown json`);}
		const response = await this.restFetch({
			method: 'POST',
			headers: {'Content-Type': 'application/json'},
			body: jsondata
		});
		return response;
	}
	async restPut(jsondata){
		if(!jsondata){throw Error(`${this.constructor.name}.restPut: unknown json`);}
		const response = await this.restFetch({
			method: 'PUT',
			headers: {'Content-Type': 'application/json'},
			body: jsondata
		});
		return response;
	}
	async restDelete(){
		const response = await this.restFetch({
			method: 'DELETE'
		});
		return response
	}
	/**
	 * Search for the JSON (this.data)
	 * @param {string} fieldName 
	 * @param {*} value 
	 * @returns 
	 */
	findField(fieldName, value){
		try{
			return this.data.find(d => d[fieldName] === value);
		}catch(e){
			return undefined;
		}
	}
}

/** toShare Common API Configuration */
class restConfiguration extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/configuration`, 'config');

		this.onUserNameClick = this.onUserNameClick.bind(this)
		this.loadedInoutEditLayer = false;
		this.linkColorCss = document.createElement('link');

		this.linkColorCss.rel = 'stylesheet';
		this.linkColorCss.type = 'text/css';
		document.head.appendChild(this.linkColorCss);
	}
	/**
	 * By passing an array containing the names of the setting items, get the setting contents in this.data
	 * @override
	 * @param {string[]} items 
	 * @returns 
	 */
	restGet(items){
		const query = new URLSearchParams({'items': items.join(",")});
		this.url = `${this.baseURL}?${query}`;
		return super.restGet();
	}
	/**
	 * Reflect the settings by passing the setting item name and setting value.
	 * @override
	 * @param {string} item 
	 * @param {*} value 
	 * @returns 
	 */
	restPut(item, value){
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'item': item,
				'value': value
			}]
		});
		return super.restPut(jsondata);
	}
	/**
	 * Get config from json(this.data)
	 * @param {string} item - config item
	 * @returns - config value
	 */
	valueFromKey(item){
		return this.data.find(d => d['item'] === item).value;
	}
	/** Reflect basic settings ('Server name', 'Color style', 'reservation', 'inoutboard') */
	async reflection(){
		try{
			await this.restGet(['Server name', 'Color style', 'reservation', 'inoutboard']);
			this.reflectionColorStyle();
			this.reflectionIndex();
			this.reflectionInoutboard();
		}catch(e){confirm(e.message);}
	}
	/** Basic settings: Reflect server name */
	reflectionIndex(){
		const servername = this.valueFromKey('Server name');
		const aHeaderIndex = document.querySelector('header p.breadcrumbs a.index');
		if(aHeaderIndex)aHeaderIndex.innerText = servername;
	}
	/** Basic settings: Reflect color style */
	reflectionColorStyle(filename){
		const cssFileName = filename ? filename : this.valueFromKey('Color style');
		const cssFileUrl = `${location.origin}/ats/${cssFileName}.css`;
		this.linkColorCss.href = cssFileUrl;
	}
	/** Basic settings: Reflect inoutboard (Username in the top right corner) */
	async reflectionInoutboard(){
		if(this.valueFromKey('inoutboard') == 'true'){
			const spanUserName = document.querySelector('.user-name');
			spanUserName.dataset.tooltip = '出退勤 行先';
			spanUserName.classList.add('attendance');
			spanUserName.addEventListener('click', this.onUserNameClick);
			try{
				if(Object.keys(atsAttendance.data).length === 0)
					await atsAttendance.restGet(atsCurrentUserId);
			}catch(e){updateUserAttendance();}
		}else{
			const spanUserName = document.querySelector('.user-name');
			delete spanUserName.dataset.tooltip;
			spanUserName.classList.remove('attendance');
			spanUserName.removeEventListener('click', this.onUserNameClick);
			atsAttendance.data =[];
		}
		updateUserAttendance();
	}
	/**
	 * @private
	 */
	async onUserNameClick(){
		if(!this.loadedInoutEditLayer){
			await this.loadInoutEditLayer();
		}else{
			showInoutEditLayer(atsCurrentUserId);	// inoutboard.js
		}
	}
	/**
	 * @private
	 */
	async loadInoutEditLayer(){
		try{
			const inoutboardCssFileUrl = `${location.origin}/ats/inoutboard.css`;
			const inoutboardHtmlFileUrl = `${location.origin}/ats/inoutboard.html`;
			const inoutboardJsFileUrl = `${location.origin}/ats/inoutboard.js`;

			const linkCssInoutboard = document.createElement('link');
			linkCssInoutboard.rel = 'stylesheet';
			linkCssInoutboard.type = 'text/css';
			linkCssInoutboard.href = inoutboardCssFileUrl;
			document.head.appendChild(linkCssInoutboard);

			await fetch(inoutboardHtmlFileUrl)
				.then(res => res.text())
				.then(data => {
					document.body.insertAdjacentHTML('beforeend', data);
				});

			const scriptInoutboard = document.createElement('script');
			scriptInoutboard.src = inoutboardJsFileUrl;
			scriptInoutboard.defer = true;
			document.head.appendChild(scriptInoutboard);
			this.loadedInoutEditLayer = true;
		}catch(e){confirm(e.message);}
	}
}

/** toShare Common API Users */
class restUsers extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/users`, 'users');
	}
	/**
	 * Get user information
	 * @override
	 * @param {number|string|undefined} userArg - Specify the target user
	 * - number: Get user information from userId
	 * - string: Get user information from userAccount
	 * - USER_ID_ALL(const undefined): Get information about all users
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 * @example
	 * .restGet()	// Get all users without hiding
	 */
	restGet(userArg, withDisable){
		if(typeof userArg === 'number'){
			const query = new URLSearchParams({'user_id': userArg});
			this.url = `${this.baseURL}?${query}`;
		}else if(typeof userArg === 'string'){
			const query = new URLSearchParams({'user_account': userArg});
			this.url = `${this.baseURL}?${query}`;
		}else if(userArg === undefined){
			if(withDisable === undefined){
				this.url = this.baseURL;
			}else{
				const query = new URLSearchParams({'with_disable': 1});
				this.url = `${this.baseURL}?${query}`;
			}
		}
		return super.restGet()
	}
	/** Add suer
	 * @override
	 */
	restPost(userAccount, userName, userDesc, userSeries){
		if(!userAccount || !userName || (userDesc===undefined) || (userSeries===undefined)){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		this.url = this.baseURL;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'user_account': userAccount,
				'user_name': userName,
				'user_desc': userDesc,
				'user_series': userSeries
			}]
		});
		return super.restPost(jsondata);
	}
	/** Edit user information
	 * @override
	 */
	restPut(userId, userAccount, userName, userDesc, userSeries, withDisable){
		if(!userId){throw Error(`${this.constructor.name}.restPut: all required argument.`);}
		const source = this.findField('user_id', userId);
		const dest = {};
		if(source){
			dest.user_id = source.user_id;
			dest.user_account = userAccount ? userAccount : source.user_code;
			dest.user_name = userName ? userName : source.user_name;
			dest.user_desc = altHCO(userDesc, source.user_desc);
			dest.user_series = altHCO(userSeries, source.user_series);
			dest.disable = withDisable;
		}else{
			if(!userId || !userAccount || !userName || (userDesc===undefined) || (userSeries===undefined)){
				throw Error(`${this.constructor.name}.restPut: all required argument.`);
			}else{
				dest.user_id = userId;
				dest.user_account = userAccount;
				dest.user_name = userName;
				dest.user_desc = userDesc;
				dest.user_series = userSeries;
				dest.user_disable = withDisable;
			}
		}
		const query = new URLSearchParams({'user_id': userId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/** Delete user
	 * @override
	 */
	restDelete(userId){
		if(!userId){throw Error(`${this.constructor.name}.restDelete: userId is a required argument.`);}
		const query = new URLSearchParams({
			'user_id': userId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/**
	 * Sort users
	 * @param {number[]} userIds
	 * @returns 
	 */
	restOrderPut(userIds){
		if(!userIds){throw Error(`${this.constructor.name}.restOrderPut: userIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:userIds});
		return super.restPut(jsondata);
	}
}

/** toShare Reservation API ReservationTables */
class restReservationTables extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/tables`, 'tables');
	};
	/**
	 * Get Reservation information
	 * @override
	 * @param {number|undefined} tableId - Reservation Table TableId
	 * - number: Get tableId Reservation Table information
	 * - TABLE_ID_ALL(const undefined): Get information about all Reservation Tables
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restGet(tableId, withDisable){
		if(tableId === undefined){
			if(withDisable === undefined){
				this.url = this.baseURL;
			}else{
				const query = new URLSearchParams({'with_disable': 1});
				this.url = `${this.baseURL}?${query}`;
			}
		}else{
			const query = new URLSearchParams({'table_id': tableId});
			this.url = `${this.baseURL}?${query}`;
		}
		return super.restGet()
	}
	/** Add Reservation Table
	 * @override
	 */
	restPost(tableName, tableCustom, tableUri, tableDesc, itemName, itemBulk){
		if(!tableName || (tableCustom===undefined) || !tableUri  || (tableDesc===undefined) || (itemName===undefined) || (itemBulk===undefined)){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		this.url = this.baseURL;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'table_name': tableName,
				'table_custom': tableCustom,
				'table_uri': tableUri,
				'table_desc': tableDesc,
				'item_name': itemName,
				'item_bulk': itemBulk
			}]
		});
		return super.restPost(jsondata);
	}
	/** Edit Raservation Table information
	 * @override
	 */
	restPut(tableId, tableName, tableCustom, tableUri, tableDesc, itemName, itemBulk, withDisable){
		if(!tableId){throw Error(`${this.constructor.name}.restPut: tableId is a required argument.`);}
		const source = this.findField('table_id', tableId);
		const dest = {};
		if(source){
			dest.table_id = source.table_id;
			dest.table_name = tableName ? tableName : source.table_name;
			dest.table_custom = (tableCustom != undefined) ? tableCustom : source.table_custom;
			dest.table_uri = tableUri ? tableUri : source.table_uri;
			dest.table_desc = altHCO(tableDesc, source.table_desc);
			dest.item_name = altHCO(itemName, source.item_name);
			dest.item_bulk = altHCO(itemBulk, source.item_bulk);
			dest.disable = (withDisable != undefined) ? withDisable : source.disable;
		}else{
			if(!tableId || !tableName || (tableCustom===undefined) || !tableUri  || (tableDesc===undefined) || (itemName===undefined) || (itemBulk===undefined)){
				throw Error(`${this.constructor.name}.restPut: required argument.`);
			}else{
				dest.table_id = tableId;
				dest.table_name = tableName;
				dest.table_custom = tableCustom;
				dest.table_uri = tableUri;
				dest.table_desc = tableDesc;
				dest.item_name = itemName;
				dest.item_bulk = itemBulk;
				dest.disable = withDisable;
			}
		}
		this.url = `${this.baseURL}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/** Delete Reservation Talbe
	 * @override
	 */
	restDelete(tableId){
		if(!tableId){throw Error(`${this.constructor.name}.restDelete: tableId is a required argument.`);}
		const query = new URLSearchParams({
			'table_id': tableId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/** Sort Reservations */
	restOrderPut(tableIds){
		if(!tableIds){throw Error(`${this.constructor.name}.restOrderPut: tableIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:tableIds});
		return super.restPut(jsondata);
	}
}

/** toShare Reservation API Items */
class restItems extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/items`, 'items');
		this.tableId = 0;
	};
	/**
	 * Get items associated with Reservations
	 * @override
	 * @param {number} tableId - Reservation tableId
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restGet(tableId, withDisable){
		this.tableId = tableId ? tableId : this.tableId;
		if(!this.tableId){throw Error(`${this.constructor.name}.restGet: tableId is not set.`);}
		const query = new URLSearchParams({'table_id': this.tableId});
		if(withDisable != undefined){query.append('with_disable',1);}
		this.url = `${this.baseURL}?${query}`;
		return super.restGet();
	}
	/** Add Item
	 * @override
	 * Requires .restGet to be executed first
	 */
	restPost(itemName, itemDesc){
		if(!this.tableId){throw Error(`${this.constructor.name}.restPost: tableId is not set.`);}
		if(!itemName || (itemDesc===undefined)){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		const query = new URLSearchParams({'table_id': this.tableId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'item_name': itemName,
				'item_desc': itemDesc
			}]
		});
		return super.restPost(jsondata);
	}
	/** Edit Item
	 * @override
	 * Requires .restGet to be executed first
	 */
	restPut(itemId, itemName, itemDesc, withDisable){
		if(!itemId){throw Error(`${this.constructor.name}.restPut: itemId is a required argument.`);}
		const source = this.findField('item_id', itemId);
		const dest = {};
		if(source){
			dest.item_id = source.item_id;
			dest.item_name = itemName ? itemName : source.item_name;
			dest.item_desc = altHCO(itemDesc, source.item_desc);
			dest.disable = withDisable;
		}else{
			if(!itemId || !itemName || !itemDesc){
				throw Error(`${this.constructor.name}.restPut: required argument.`);
			}else{
				dest.item_id = itemId;
				dest.item_name = itemName;
				dest.item_desc = itemDesc;
				dest.disable = withDisable;
			}
		}
		const query = new URLSearchParams({'table_id': this.tableId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/** Delete Item
	 * @override
	 * Requires .restGet to be executed first
	 */
	restDelete(itemId){
		if(!this.tableId){throw Error(`${this.constructor.name}.restDelete: tableId is not set.`);}
		if(!itemId){throw Error(`${this.constructor.name}.restDelete: itemId is a required argument.`);}
		const query = new URLSearchParams({
			'table_id': this.tableId,
			'item_id': itemId
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/** Sort Items */
	restOrderPut(itemIds){
		if(!itemIds){throw Error(`${this.constructor.name}.restOrderPut: itemIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:itemIds});
		return super.restPut(jsondata);
	}
}

/** toShare Reservation API Calendar */
class restCalendar extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/calendar`, 'calendar');
		this.startDate = new Date();
		this.endDate = new Date();
	}
	/**
	 * Get calendar information from start date to end date
	 * @override
	 * @param {Date} startDate 
	 * @param {Date} endDate 
	 * @returns 
	 */
	restGet(startDate, endDate){
		this.startDate.setTime(
			startDate ? startDate.getTime() : this.startDate.getTime()
		);
		this.endDate.setTime(
			endDate ? endDate.getTime() : this.endDate.getTime()
		);
		const query = new URLSearchParams({
			'start_date': formatDate(this.startDate, 'yyyy-MM-dd'),
			'end_date': formatDate(this.endDate, 'yyyy-MM-dd')
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restGet();
	}
	/**
	 * Set calendar information from the start date to the end date
	 * @override
	 * @param {Date} startDate 
	 * @param {Date} endDate 
	 * @param {object} holData - json(this.data schema)
	 * @returns 
	 */
	restPut(startDate, endDate, holData){
		this.startDate.setTime(
			startDate ? startDate.getTime() : this.startDate.getTime()
		);
		this.endDate.setTime(
			endDate ? endDate.getTime() : this.endDate.getTime()
		);
		this.data = holData ? holData : this.data;
		const query = new URLSearchParams({
			'start_date': formatDate(this.startDate, 'yyyy-MM-dd'),
			'end_date': formatDate(this.endDate, 'yyyy-MM-dd')
		});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:this.data});
		return super.restPut(jsondata);
	}
	/**
	 * Returns whether the target day is a weekday, a legal holiday, or a scheduled holiday
	 * @param {Date} targetDate 
	 * @returns {string|undefined} - day type
	 * - undefined: weekday
	 * - 'lh': legal holiday
	 * - 'sh': scheduled holiday
	 */
	dateType(targetDate){
		// return = undefined(平日扱) || lh(legal holiday:法定休日) || sh(schedule holiday:所定休日)
		const dateData = this.findField('date', formatDate(targetDate, 'yyyy-MM-dd'));
		if(dateData !== undefined){					// ユーザー定義がある場合、
			if(dateData.type === 'sh'){return 'sh';} // 所定休日
		}else{										// ユーザー定義が無い場合、
			if(targetDate.getDay()===0){return 'lh';} // 日曜を法定休日
			if(targetDate.getDay()===6){return 'sh';} // 土曜を所定休日
		}
	}
}

/** toShare Reservation API Reservations */
class restReservations extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/reservations`, 'reservations');
		this.tableId = 0;
		this.startDateTime = new Date();
		this.endDateTime = new Date();
	}
	/**
	 * Get resebations from start date to end date
	 * @override
	 * @param {number} tableId - Reservation table tableId
	 * @param {Date} startDateTime 
	 * @param {Date} endDateTime 
	 * @returns 
	 */
	restGet(tableId, startDateTime, endDateTime){
		this.tableId = tableId ? tableId : this.tableId;
		this.startDateTime.setTime(
			startDateTime ? startDateTime.getTime() : this.startDateTime.getTime()
		);
		this.endDateTime.setTime(
			endDateTime ? endDateTime.getTime() : this.endDateTime.getTime()
		);
		if(!this.tableId){throw Error(`${this.constructor.name}.restGet: tableId is not set.`);}
		const query = new URLSearchParams({
			'table_id': this.tableId,
			'start_datetime': this.startDateTime.toISOString(),
			'end_datetime': this.endDateTime.toISOString()
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restGet();
	}
	/**
	 * Get the last reservation information up to the specified date
	 * @param {number} tableId 
	 * @param {Date} dateTime 
	 * @returns 
	 */
	restGetLastUses(tableId, dateTime){
		this.tableId = tableId ? tableId : this.tableId;
		this.endDateTime.setTime(
			dateTime ? dateTime.getTime() : this.endDateTime.getTime()
		);
		if(!this.tableId){throw Error(`${this.constructor.name}.restGetLastUses: tableId is not set.`);}
		const query = new URLSearchParams({
			'table_id': this.tableId,
			'datetime': this.endDateTime.toISOString(),
			'last_uses': 1
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restGet();
	}
	/**
	 * Set Reservation
	 * @override
	 * @param {number} itemIds
	 * @param {number} userId
	 * @param {Date} startDateTime
	 * @param {Date} endDateTime
	 * @param {string} desc
	 * @returns 
	 */
	restPost(itemIds, userId, startDateTime, endDateTime, desc){
		if(!this.tableId){throw Error(`${this.constructor.name}.restPost: tableId is not set.`);}
		if(!itemIds || !userId || !startDateTime || !endDateTime || !desc){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		const query = new URLSearchParams({'table_id': this.tableId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'item_ids': itemIds,
				'user_id': userId,
				'start_datetime': startDateTime.toISOString(),
				'end_datetime': endDateTime.toISOString(),
				'reservation_desc': desc
			}]
		});
		return super.restPost(jsondata);
	}
	/**
	 * Edit Reservation information
	 * @override
	 * @param {number} reservationId
	 * @param {number} itemIds
	 * @param {number} userId
	 * @param {Date} startDateTime
	 * @param {Date} endDateTime
	 * @param {string} desc
	 * @returns 
	 */
	restPut(reservationId, itemIds, userId, startDateTime, endDateTime, desc){
		if(!reservationId){throw Error(`${this.constructor.name}.restPut: reservationId is a required argument.`);}
		const source = this.findField('reservation_id', reservationId);
		const dest = {};
		if(source){
			dest.reservation_id = source.reservation_id;
			dest.item_ids = itemIds ? itemIds : source.item_ids;
			dest.user_id = userId ? userId : source.User_id; 
			dest.start_datetime = startDateTime ? startDateTime.toISOString() : source.start_dateTime;
			dest.end_datetime = endDateTime ? endDateTime.toISOString() : source.end_dateTime;
			dest.reservation_desc = altHCO(desc, source.desc);
		}else{
			if(!itemIds || !userId || !startDateTime || !endDateTime){
				throw Error(`${this.constructor.name}.restPut: required argument.`);
			}else{
				dest.reservation_id = reservationId;
				dest.item_ids = itemIds;
				dest.user_id = userId;
				dest.start_datetime = startDateTime.toISOString();
				dest.end_datetime = endDateTime.toISOString();
				dest.reservation_desc = desc;
			}
		}
		const query = new URLSearchParams({'table_id': this.tableId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/**
	 * Delete Reservation 
	 * @override
	 * @param {number} reservationId 
	 * @returns 
	 */
	restDelete(reservationId){
		if(!this.tableId){throw Error(`${this.constructor.name}.restDelete: tableId is not set.`);}
		if(!reservationId){throw Error(`${this.constructor.name}.restDelete: reservationId is a required argument.`);}
		const query = new URLSearchParams({
			'table_id': this.tableId,
			'reservation_id': reservationId
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
};

/** toShare Inoutboard API Inoutboards */
class restInoutboards extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/inoutboards`, 'inoutboards');
	};
	/**
	 * Get Inoutboards
	 * @override
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restGet(withDisable){
		if(withDisable === undefined){
			this.url = this.baseURL;
		}else{
			const query = new URLSearchParams({'with_disable': 1});
			this.url = `${this.baseURL}?${query}`;
		}
		return super.restGet()
	}
	/** Set Inoutboard
	 * @override
	 */
	restPost(boardName){
		if(!boardName){throw Error(`${this.constructor.name}.restPost: boardName required argument.`);}
		this.url = this.baseURL;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'board_name': boardName
			}]
		});
		return super.restPost(jsondata);
	}
	/** Edit Innoutboard information
	 * @override
	 */
	restPut(boardId, boardName, withDisable){
		if(!boardId){throw Error(`${this.constructor.name}.restPut: boardId is a required argument.`);}
		const source = this.findField('board_id', boardId);
		const dest = {};
		if(source){
			dest.board_id = source.board_id;
			dest.board_name = boardName ? boardName : source.board_name;
			dest.disable = (withDisable != undefined) ? withDisable : source.disable;
		}else{
			if(!boardId || !boardName){
				throw Error(`${this.constructor.name}.restPut: required argument.`);
			}else{
				dest.board_id = boardId;
				dest.board_name = boardName;
				dest.disable = withDisable;
			}
		}
		this.url = `${this.baseURL}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/** Delete Inoutoboard
	 * @override
	 */
	restDelete(boardId){
		if(!boardId){throw Error(`${this.constructor.name}.restDelete: boardId is a required argument.`);}
		const query = new URLSearchParams({
			'board_id': boardId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/** Sort Inoutboards */
	restOrderPut(boardIds){
		if(!boardIds){throw Error(`${this.constructor.name}.restOrderPut: boardIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:boardIds});
		return super.restPut(jsondata);
	}
}

/** toShare Inoutboard API Whereabouts */
class restWhereabouts extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/whereabouts`, 'whereabouts');
	};
	/**
	 * Get Whereabouts
	 * @override
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restGet(withDisable){
		if(withDisable === undefined){
			this.url = this.baseURL;
		}else{
			const query = new URLSearchParams({'with_disable': 1});
			this.url = `${this.baseURL}?${query}`;
		}
		return super.restGet()
	}
	/**
	 * Set whereabouts
	 * @override
	 * @param {string} whereabouts - whereabouts item name
	 * @param {number} optGroup - Whether to treat items as a group
	 * - 0:treat as a item
	 * - 1:treat as a group
	 * @returns 
	 */
	restPost(whereabouts, optGroup){
		if(!whereabouts){throw Error(`${this.constructor.name}.restPost: whereabouts required argument.`);}
		this.url = this.baseURL;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'whereabouts': whereabouts,
				'optGroup': optGroup
			}]
		});
		return super.restPost(jsondata);
	}
	/**
	 * Edit whereabouts information
	 * @override
	 * @param {number} whereaboutsId 
	 * @param {string} whereabouts 
	 * @param {number} optGroup 
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restPut(whereaboutsId, whereabouts, optGroup, withDisable ){
		if(!whereaboutsId){throw Error(`${this.constructor.name}.restPut: whereaboutsId is a required argument.`);}
		const source = this.findField('whereabouts_id', whereaboutsId);
		const dest = {};
		if(source){
			dest.whereabouts_id = source.whereabouts_id;
			dest.whereabouts = whereabouts ? whereabouts : source.whereabouts;
			dest.disable = (withDisable != undefined) ? withDisable : source.disable;
			dest.optGroup = (optGroup !=  undefined) ? optGroup : source.optGroup;
		}else{
			if(!whereaboutsId || !whereabouts){
				throw Error(`${this.constructor.name}.restPut: required argument.`);
			}else{
				dest.whereabouts_id = whereaboutsId;
				dest.whereabouts = whereabouts;
				dest.disable = withDisable;
				dest.optGroup = optGroup;
			}
		}
		this.url = `${this.baseURL}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/**
	 * Delete whereabouts
	 * @override
	 * @param {number} whereaboutsId 
	 * @returns 
	 */
	restDelete(whereaboutsId){
		if(!whereaboutsId){throw Error(`${this.constructor.name}.restDelete: whereaboutsId is a required argument.`);}
		const query = new URLSearchParams({
			'whereabouts_id': whereaboutsId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/** Sort whereabouts */
	restOrderPut(whereaboutsIds){
		if(!whereaboutsIds){throw Error(`${this.constructor.name}.restOrderPut: whereaboutsIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:whereaboutsIds});
		return super.restPut(jsondata);
	}
}

/** toShare Inoutboard API Attendance */
class restAttendance extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/attendance`, 'attendance');
	}
	/**
	 * Get Attendance
	 * @override
	 * @param {number|undefined} userId - Specify the target user
	 * - number: Get user attendance from userId
	 * - USER_ID_ALL(const undefined): Get attendance about all users
	 * @param {number|undefined}  [withDisable] - WITHDISABLED
	 * - WITHDISABLED(const 1): Include Hide
	 * - undefined(Not specified): Does not include hiding
	 * @returns 
	 */
	restGet(userId, withDisable){
		if(userId === undefined){
			if(withDisable === undefined){
				this.url = this.baseURL;
			}else{
				const query = new URLSearchParams({'with_disable': 1});
				this.url = `${this.baseURL}?${query}`;
			}
		}else{
			const query = new URLSearchParams({'user_id': userId});
			this.url = `${this.baseURL}?${query}`;
		}
		return super.restGet()
	}
	/**
	 * Add or modify a user's affiliation
	 * @override
	 * @param {number[]} userIds 
	 * @param {number} boardId 
	 * @returns 
	 */
	restPost(userIds, boardId){
		if(!userIds || !boardId){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		this.url = this.baseURL;
		var arrayUsers = [];
		for(const userId of userIds){
			arrayUsers.push({'user_id': userId, 'board_id': boardId});
		}
		const jsondata = JSON.stringify({[this.jsonRoot]:arrayUsers});
		return super.restPost(jsondata);
	}
	/**
	 * Set user attendance status
	 * @override
	 * @param {number} userId 
	 * @param {number} clockStatus - clock status
	 * - 1: clock in
	 * - 0: clock out
	 * @param {string} whereabouts 
	 * @param {string} description 
	 * @returns 
	 */
	restPut(userId, clockStatus, whereabouts, description){
		if(!userId){throw Error(`${this.constructor.name}.restPut: all required argument.`);}
		const source = this.findField('user_id', userId);
		const dest = {};
		if(source){
			dest.user_id = source.user_id;
			dest.clock_status = altHCO(clockStatus, source.clock_status);
			dest.whereabouts = altHCO(whereabouts, source.whereabouts);
			dest.description = altHCO(description, source.description);
		}else{
			if(!userId || !clockStatus || (whereabouts===undefined) || (description===undefined)){
				throw Error(`${this.constructor.name}.restPut: all required argument.`);
			}else{
				dest.user_id = userId;
				dest.clock_status = clockStatus;
				dest.whereabouts = whereabouts;
				dest.description = description;
			}
		}
		this.url = `${this.baseURL}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/**
	 * Remove a user from a group
	 * @override
	 * @param {number} userId 
	 * @returns 
	 */
	restDelete(userId){
		if(!userId){throw Error(`${this.constructor.name}.restDelete: userId is a required argument.`);}
		const query = new URLSearchParams({
			'user_id': userId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
}

/**
 * Updated display of signed-in users
 */
function updateUserAttendance(){
	const userName = document.querySelector('.user-name');
	const atsUser = atsAttendance.findField('user_id', atsCurrentUserId);
	if(atsUser){
		switch(atsUser.clock_status){
			case 0:
				userName.classList.add('clock-out');
				userName.classList.remove('clock-in');
				userName.classList.remove('default');
				break;
			case 1:
				userName.classList.add('clock-in');
				userName.classList.remove('clock-out');
				userName.classList.remove('default');
				break;
		}
	}else{
		userName.classList.add('default');
		userName.classList.remove('clock-in');
		userName.classList.remove('clock-out');
	}
}

/**
 * toShare Inoutboard API Co Hosts Regist
 */
class restCoHosts extends restClient {
	/**
	 * @override
	 */
	constructor(){
		super(`${basePath}/co_hosts`, 'co_hosts');
	}
	/**
	 * Get Co Host List
	 * @override
	 * @param {number} withUrl - Choose whether to get the URL
	 * - WITH_URL(const 1): Get the URL
	 * - undefined: Do not retrieve URLs
	 * @returns 
	 */
	restGet(withUrl){
		if(withUrl === undefined){
			this.url = this.baseURL;
		}else{
			const query = new URLSearchParams({'with_url': 1});
			this.url = `${this.baseURL}?${query}`;
		}
		return super.restGet()
	}
	/**
	 * Add Co Host
	 * @override
	 * @param {string} hostName
	 * @param {string} hostHost - Co Host IP Address
	 * @param {number} hostPort
	 * @returns 
	 */
	restPost(hostName, hostHost, hostPort){
		if(!hostName || !hostHost || !hostPort){
			throw Error(`${this.constructor.name}.restPost: all required argument.`);
		}
		this.url = this.baseURL;
		const jsondata = JSON.stringify({
			[this.jsonRoot]:[{
				'host_name': hostName,
				'host_host': hostHost,
				'host_port': hostPort
			}]
		});
		return super.restPost(jsondata);
	}
	/**
	 * Edit Co Host
	 * @override
	 * @param {number} hostId 
	 * @param {string} hostName 
	 * @param {string} hostHost - Co Host IP Address 
	 * @param {number} hostPort 
	 * @returns 
	 */
	restPut(hostId, hostName, hostHost, hostPort){
		if(!hostId){throw Error(`${this.constructor.name}.restPut: all required argument.`);}
		const source = this.findField('host_id', hostId);
		const dest = {};
		if(source){
			dest.host_id = source.host_id;
			dest.host_name = hostName ? hostName : source.user_name;
			dest.host_host = hostHost ? hostHost : source.host_host;
			dest.host_port = hostPort ? hostPort : source.host_port;
		}else{
			if(!hostId || !hostName || !hostHost || !hostPort){
				throw Error(`${this.constructor.name}.restPut: all required argument.`);
			}else{
				dest.host_id = hostId;
				dest.host_name = hostName;
				dest.host_host = hostHost;
				dest.host_port = hostPort;
			}
		}
		const query = new URLSearchParams({'host_id': hostId});
		this.url = `${this.baseURL}?${query}`;
		const jsondata = JSON.stringify({[this.jsonRoot]:[dest]});
		return super.restPut(jsondata);
	}
	/**
	 * Delete Co Host
	 * @override
	 * @param {number} hostId 
	 * @returns 
	 */
	restDelete(hostId){
		if(!hostId){throw Error(`${this.constructor.name}.restDelete: hostId is a required argument.`);}
		const query = new URLSearchParams({
			'host_id': hostId,
		});
		this.url = `${this.baseURL}?${query}`;
		return super.restDelete();
	};
	/** Sort Co Host */
	restOrderPut(hostIds){
		if(!hostIds){throw Error(`${this.constructor.name}.restOrderPut: hostIds is a required argument.`);}
		this.url = `${this.baseURL}/order`;
		const jsondata = JSON.stringify({[this.jsonRoot]:hostIds});
		return super.restPut(jsondata);
	}
}

const TABLE_ID_ALL = undefined;
const USER_ID_ALL = undefined;
const WITHDISABLED = 1;
const WITH_URL = 1;
/** An instance of restConfiguration (Atelier Arika toShare Common API) */
const atsConfiguration = new restConfiguration();
/** An instance of restUsers (Atelier Arika toShare Common API) */
const atsUsersTable = new restUsers();
/** An instance of restReservationTables(Atelier Arika toShare Reservation API) */
const atsReservationTables = new restReservationTables();
/** An instance of restItems (Atelier Arika toShare Reservation API) */
const atsItems = new restItems();
/** An instance of restReserbations (Atelier Arika toShare Reservation API) */
const atsReservations = new restReservations();
/** An instance of restCalendar (Atelier Arika toShare Reservation API) */
const atsCalendar = new restCalendar();
/** An instance of restWhereabouts (Atelier Arika toShare Inoutboad API) */
const atsWhereabouts = new restWhereabouts();
/** An instance of restInoutboads (Atelier Arika toShare Inoutboad API) */
const atsInoutboards = new restInoutboards();
/** An instance of restAttendance (Atelier Arika toShare Inoutboad API) */
const atsAttendance = new restAttendance();
/** An instance of restCoHosts (Atelier Arika toShare CoHosts API) */
const atsCoHosts = new restCoHosts();
